/*
 * Copyright (C) 2004  Heinz Mauelshagen, Red Hat GmbH. All rights reserved.
 *
 * See file LICENSE at the top of this source tree for license information.
 */

/*
 * Template to implement metadata format handlers.
 */

#include "internal.h"
#define	FORMAT_HANDLER
#include "template.h"
#include <datastruct/byteorder.h>

static const char *handler = "template";

/* Make up RAID device name. */
/* CODEME: implement creation of senseful name for the RAID device */
static size_t __name(struct template *template, char *str, size_t len,
		     unsigned char type)
{
	return snprintf(str, len, "template") + 1;
}

static char *_name(struct template *template, unsigned char type)
{
	size_t len;
	char *ret;

	len = __name(template, NULL, 0, type);
	if ((ret = dbg_malloc(len + 1))) {
		__name(template, ret, len, type);
		mk_alpha(ret, strlen(ret));
	} else
		log_err(lc, "%s: failed allocating name string", handler);

	return ret;
}

/* Mapping of template types to generic types */
/*
 * CODEME: mappings of template private level types to generic ones 
 * (see metadata.h for generic ones)
 */
static struct types types[] = {
	{ TEMPLATE_T_SPAN,  t_linear },
	{ TEMPLATE_T_RAID0, t_raid0 },
        { 0, t_undef}
};

/* Neutralize disk type using generic metadata type mapping function */
static enum type template_type(struct lib_context *lc, struct dmraid_dev *rd)
{
	return rd_type(types,
		       (unsigned int) ((struct template*) rd->meta_areas[0].area)->type);
}

/*
 * Read a Template RAID device
 */
/* CODEME: endianess conversion */
#if	BYTE_ORDER == LITTLE_ENDIAN
#  define	to_cpu(x)
#else
static void to_cpu(struct template *template)
{
	template->something = le32_to_cpu(template->something);
	...
}
#endif

static int setup_rd(struct lib_context *lc, struct dmraid_dev *rd,
		    struct template *template, struct disk_info *di);
static struct dmraid_dev *template_read(struct lib_context *lc,
					struct disk_info *di)
{
	struct template *template;
	struct dmraid_dev *rd = NULL;

	/* Allocate space for private ondisk metadata */
	if (!(template = alloc_private(lc, "template", sizeof(*template))))
		goto out;

	/* CODEME: calculate TEMPLATE_CONFIGOFFSET correctly */
	if (!read_file(lc, di->path, (void *) template,
		       sizeof(*template), TEMPLATE_CONFIGOFFSET)
		goto bad;

	to_cpu(template);

	/* CODEME: unique identification of template metadata */
	if (strncmp(template->magic_name, TEMPLATE_MAGIC_NAME,
		    sizeof(TEMPLATE_MAGIC_NAME) - 1))
		goto bad;
	
	/* File metadata for tests (writes ${devicename}.dat file) */
	file_metadata(lc, handler, di->path, template, sizeof(*template),
		      TEMPLATE_CONFIGOFFSET);

	/* Allocate RAID device structure (see metadata.c) */
	if (!(rd = alloc_dev(lc)))
		goto bad;

	/*
	 * Setup generic RAID device structure members using
	 * ondisk inforation
	 */
	if (setup_rd(lc, rd, template, di))
		goto out;

	free_dev(lc, rd);
	rd = NULL;

   bad:
	dbg_free(template);

   out:
	return rd;
}

/* Add a Template RAID device to a set */
static struct dmraid_set *template_group(struct lib_context *lc,
					 struct dmraid_dev *rd)
{
	/* CODEME: add grouping logic */
	log_err(lc, "%s: implement grouping logic for RAID set", handler);
	return NULL;
}

/* CODEME: Write private RAID metadata to device */
static int template_write(struct lib_context *lc,
			  struct dmraid_dev *rd, int erase)
{
	int ret;
#if	BYTE_ORDER != LITTLE_ENDIAN
	struct template *template = rd->meta_areas[0].area;
#endif

	/* CODEME: in case there's more complex metadata areas */
	to_disk(template);
	ret = write_metadata(lc, handler, rd, -1, erase);
	to_cpu(template);
	return ret;
}

/*
 * Check integrity of a RAID set.
 */
static int template_check(struct lib_context *lc, struct dmraid_set *rs)
{
	/* CODEME: implement */
	log_err(lc, "%s: implement RAID set integrity checks", handler);
	return 0;
}

#ifdef DMRAID_NATIVE_LOG
/*
 * Log native information about the RAID device.
 */
static void template_log(struct lib_context *lc, struct dmraid_dev *rd)
{
	struct template *template = (struct template*) rd->meta_areas[0].area;

	/* CODEME: implement (use P(), ... macors from format.h */
	log_print(lc, "%s: implement displaying metadata variables", handler);
}
#endif

static struct dmraid_format template_format = {
	.name	= "template",
	.descr	= "(-) Template RAID",
	.format = FMT_RAID,
	.read	= template_read,
	.write	= template_write,
	.group	= template_group,
	.check	= template_check,
#ifdef DMRAID_NATIVE_LOG
	.log	= template_log,
#endif
	.type	= template_type,
};

/* Register this format handler with the format core */
int register_template(struct lib_context *lc)
{
	return register_format_handler(lc, &template_format);
}

/* CODEME: Set the RAID device contents up derived from the TEMPLATE ones */
static int setup_rd(struct lib_context *lc, struct dmraid_dev *rd,
		    struct template *template, struct disk_info *di)
{
        rd->di = di;
        rd->name = _name(template, 0);
	rd->fmt = &template_format;

	rd->offset = TEMPLATE_DATAOFFSET;
	rd->sectors = di->sectors - 1; /* CODEME: correct sectors */

	if (!(rd->meta_areas = alloc_meta_areas(lc, rd, handler, 1)))
		return 0;

	rd->meta_areas[0].offset = TEMPLATE_CONFIGOFFSET >> 9;
	rd->meta_areas[0].size = sizeof(*template);
	rd->meta_areas[0].area = (void *) template;

	rd->status = s_operational; /* CODEME: derive from metadata. */

	return 1;
}
